#
# (c) 1998 Michel Beaudouin-Lafon, mbl@daimi.au.dk
#
# small example of aligning instrument
#
# ### a faire: mettre des icones dans la palette ###

source sm.tcl

wm title . "Magnetic Guidelines"

canvas .c -width 640 -height 480
pack .c -expand on -fill both

proc lset {lvars lvals} {
	uplevel [list foreach $lvars $lvals {break}]
}

# ---- creating places and transitions

proc CreateTrans {x y} {
	set hsize 20
	set vsize 10
	set item [.c create rect [expr $x - $hsize] [expr $y - $vsize] [expr $x + $hsize] [expr $y + $vsize] \
		-outline blue -fill white -width 2 -tag {trans item}]
	return $item
}

proc CreatePlace {x y} {
	set hsize 20
	set vsize 10
	set item [.c create oval [expr $x - $hsize] [expr $y - $vsize] [expr $x + $hsize] [expr $y + $vsize] \
		-outline blue -fill white -width 2 -tag {trans item}]
	return $item
}

# ---- Creating guidelines

proc CreateHGuide {x y} {
	set item [.c create line 0 $y 1000 $y -width 3 -fill lightgreen -tag {guide hguide}]
	.c addtag g$item withtag $item
	.c lower $item
	return $item
}

proc CreateVGuide {x y} {
	set item [.c create line $x 0 $x 1000 -width 3 -fill lightgreen -tag {guide vguide}]
	.c addtag g$item withtag $item
	.c lower $item
	return $item
}

# ---- Menu bar

.c create rectangle 3 3 638 23 -fill lightgrey -tag menubar
.c create rectangle 300 4 380 22 -outline {} -fill lightgrey -tag menutitle

set HilitedItem ""

proc CreateMenu {} {
	global HilitedItem
	
	.c create rectangle 300 22 380 223 -fill lightgrey -tag menu
	for {set y 23} {$y < 223} {incr y 20} {
		.c create rectangle 301 $y 379 [expr $y + 18] -tag {menu menuitem} \
			-outline lightgrey -fill lightgrey
	}
	set HilitedItem ""
}

proc RemoveMenu {} {
	.c delete menu
	.c itemconfig menutitle -fill lightgrey
}

proc HiliteItem {item} {
	global HilitedItem ""
	if {$HilitedItem != ""} {
		.c itemconfig $HilitedItem -fill lightgrey
		set HilitedItem ""
	}
	if {$item != ""} {
		set HilitedItem $item
		.c itemconfig $HilitedItem -fill darkgrey
	}
}

# ---- Box within the canvas, with primitive window management

proc Box {path {title ""}} {

	set path .c$path
	if [winfo exist $path] {
		return ""
	}
	
	frame $path -bg black
	place $path -x 300 -y 220
	
	frame $path.box
	frame $path.close -bg lightgrey -width 15 -height 15
	label $path.title -text $title -fg white -bg black
	
	pack $path.box -side bottom -expand on -fill both -padx 2 -pady 2
	pack $path.close -side left -padx 6 -pady 2
	pack $path.title -side top -padx 2 -pady 2 -fill x
	
	StateMachine::enable closeBox $path.close
	StateMachine::enable dragBox $path
	StateMachine::enable dragBox $path.title
	
	return $path.box
}

proc CloseBox {path} {
	destroy [winfo parent $path]
}

StateMachine::define closeBox {
	state Start {
		when Down do {
			%W config -bg red
		} -> WaitUp
	}
	
	state WaitUp {
		when Leave do {
			%W config -bg lightgrey
		} -> Out
		when Up do {
			CloseBox %W
		} -> Start
	}

	state Out {
		when Enter do {
			%W config -bg red
		} -> WaitUp
		when Up do {
		} -> Start
	}
}

StateMachine::define dragBox {
	local P
	local box
	local boxpos
	
	state Start {
		when Down do {
			set P {%X %Y}
			set box %W
			while {[winfo parent $box]  != ".c"} {
				set box [winfo parent $box]
			}
			array set info [place info $box]
			set boxpos [list $info(-x) $info(-y)]
		} -> Drag
	}
	
	state Drag {
		when Move do {
			lset {x y} $P
			set dx [expr %X - $x]
			set dy [expr %Y - $y]
			lset {x y} $boxpos
			place config $box -x [expr $x + $dx] -y [expr $y + $dy]
		}
		when Up do {
		} -> Start
	}
}

# ---- toolbox

set tool "place"

proc CreateToolbox {} {
	set box [Box .box "Tools"]
	if {$box == ""} {
		return
	}
	
	global tool
	foreach t {place trans hguide vguide} {
		radiobutton $box.$t -text $t -variable tool -value $t
		pack $box.$t -side top -fill x
	}
}

CreateToolbox

# ---- Creating objects

set itemPos {}

proc StartCreate {pos} {
	global itemPos tool
	set itemPos $pos
	switch $tool {
		place {
			set newItem [eval CreatePlace $pos]
			StartDragItem $newItem $pos
		}
		trans {
			set newItem [eval CreateTrans $pos]
			StartDragItem $newItem $pos
		}
		hguide {
			set newItem [eval CreateHGuide $pos]
			StartDragHGuide $newItem $pos
		}
		vguide {
			set newItem [eval CreateVGuide $pos]
			StartDragVGuide $newItem $pos
		}
	}
}

proc DoCreate {pos} {
	global itemPos tool
	switch $tool {
		place -
		trans {
			DoDragItem $pos
		}
		hguide -
		vguide {
			DoDragGuide $pos
		}
	}
}

proc StopCreate {pos} {
	global tool
	switch $tool {
		place -
		trans {
			StopDragItem $pos
		}
		hguide -
		vguide {
			StopDragGuide $pos
		}
	}
}

# ---- Dragging items

set mouse {0 0}
set snapOffset {0 0}

proc StartDragItem {item pos} {
	global itemPos mouse
	set mouse $pos
	.c addtag drag withtag $item
	lset {x1 y1 x2 y2} [.c coords drag]
	set x [expr ($x1 + $x2) / 2]
	set y [expr ($y1 + $y2) / 2]
	set itemPos [list $x $y]
	.c itemconfig drag -fill blue
}

proc DoDragItem {pos} {
	global mouse itemPos snapOffset
	eval .c move drag $snapOffset
	
	lset {x y} $mouse
	lset {px py} $pos
	set dx [expr $px - $x]
	set dy [expr $py - $y]
	set mouse $pos
	
	lset {x y} $itemPos
	set x [expr $x + $dx]
	set y [expr $y + $dy]
	set itemPos [list $x $y]
	.c move drag $dx $dy
	
	# snap to guideline
	set dmin 10
	set hguide ""
	foreach guide [.c find withtag hguide] {
		# get guideline coordinate
		set g [lindex [.c coords $guide] 1]
		# distance from object to guideline
		set d [expr abs($g - $y)]
		# decide whether to snap
		if {$d < $dmin} {
			set dmin $d
			set gy $g
			set hguide $guide
		}
	}
	
	# same thing for vertical guidelines
	set dmin 10
	set vguide ""
	foreach guide [.c find withtag vguide] {
		set g [lindex [.c coords $guide] 0]
		set d [expr abs($g - $x)]
		if {$d < $dmin} {
			set dmin $d
			set gx $g
			set vguide $guide
		}
	}
	
	foreach guide [.c find withtag guide] {
		.c itemconfig $guide -fill lightgreen
	}
	set tags {drag item}
	set dx 0
	set dy 0
	if {$hguide != ""} {
		lappend tags g$hguide
		set dy [expr $gy - $y]
		.c itemconfig $hguide -fill green
	}
	if {$vguide != ""} {
		lappend tags g$vguide
		set dx [expr $gx - $x]
		.c itemconfig $vguide -fill green
	}
	.c itemconfig drag -tag $tags
	.c move drag $dx $dy
	
	set snapOffset [list [expr -$dx] [expr -$dy]]
}

proc StopDragItem {pos} {
	.c itemconfig drag -fill white
	.c dtag drag
	global snapOffset
	set snapOffset {0 0}
	foreach guide [.c find withtag guide] {
		.c itemconfig $guide -fill lightgreen
	}
}

# ---- Dragging items

set dragGuide ""
set dragType ""

proc StartDragHGuide {guide pos} {
	global mouse dragGuide dragType
	set dragGuide $guide
	set dragType hguide
	set mouse $pos
	.c itemconfig $guide -fill green
}

proc StartDragVGuide {guide pos} {
	global mouse dragGuide dragType
	set dragGuide $guide
	set dragType vguide
	set mouse $pos
	.c itemconfig $guide -fill green
}

proc DoDragGuide {pos} {
	global mouse dragGuide dragType
	
	lset {x y} $mouse
	lset {px py} $pos
	set dx [expr $px - $x]
	set dy [expr $py - $y]
	set mouse $pos
	switch $dragType {
		hguide { set dx 0}
		vguide { set dy 0}
	}
	.c move g$dragGuide $dx $dy
}

proc StopDragGuide {pos} {
	global dragGuide
	.c itemconfig $dragGuide -fill lightgreen
}

# ---- State machine

StateMachine::define Canvas {
	state Start {
		# moving a text item
		when Down at p onItem i withTag item do {
			StartDragItem $i $p
		} -> DragItem
		
		# moving a guideline
		when Down at p onItem i withTag hguide do {
			StartDragHGuide $i $p
		} -> DragGuide
		
		when Down at p onItem i withTag vguide do {
			StartDragVGuide $i $p
		} -> DragGuide
		
		# using the menubar
		when Enter onItem i withTag menutitle do {
			.c itemconfig menutitle -fill darkgrey
		}
		when Leave onItem i withTag menutitle do {
			.c itemconfig menutitle -fill lightgrey
		}
		when Down onItem i withTag menutitle do {
			# nothing - just catch event
		}
		when Up onItem i withTag menutitle do {
			CreateMenu
		} -> Menu
		
		# clicking on the background
		when Down at p do {
			StartCreate $p
		} -> Create
	}
	
	# drag item 
	state DragItem {
		when Move at p do {
			DoDragItem $p
		}
		when Up at p do {
			StopDragItem $p
		} -> Start
	}
	
	# drag guideline 
	state DragGuide {
		when Move at p do {
			DoDragGuide $p
		}
		when Up at p do {
			StopDragGuide $p
		} -> Start
	}
	
	# create item or guideline
	state Create {
		when Move at p do {
			DoCreate $p
		}
		when Up at p do {
			StopCreate $p
		} -> Start
	}
	
	# select item in menu
	state Menu {
		when Enter onItem i withTag menuitem do {
			HiliteItem $i
		}
		when Leave onItem i withTag menuitem do {
			HiliteItem ""
		}
		when Up do {
			RemoveMenu
		} -> Start
	}
}

StateMachine::enable Canvas .c
