#
# (c) 1998 Michel Beaudouin-Lafon, mbl@daimi.au.dk
#
# small example of indirect/direct manipulation
#

source sm.tcl

wm title . "(In)direct Manipulation"

canvas .c -width 640 -height 480
pack .c -expand on -fill both

font create f1 -family Helvetica -size 24 -weight bold
font create f2 -family Helvetica -size 24 -weight bold
set fontsize 24

.c create text 100 100 -text "Hello" -anchor nw -tag text -font f1 -fill blue
.c create text 400 200 -text "World" -anchor nw -tag text -font f2 -fill blue

proc lset {lvars lvals} {
	uplevel [list foreach $lvars $lvals {break}]
}

# ---- Handles and Selection
proc CreateHandle {x y tag} {
	set hsize 3
	.c create rectangle [expr $x - $hsize] [expr $y - $hsize]  [expr $x + $hsize] [expr $y + $hsize] \
		-fill black -tags [list handle $tag]
}

proc CreateHandles {item} {
	lset {x1 y1 x2 y2} [.c bbox $item]
	CreateHandle $x1 $y1 nw
	CreateHandle $x1 $y2 sw
	CreateHandle $x2 $y1 ne
	CreateHandle $x2 $y2 se
}

proc RemoveHandles {} {
	.c delete handle
}

proc Select {tagorid} {
	set item [.c find withtag $tagorid]
	CreateHandles $item
	.c addtag selection withtag $item
	
	# set fontsize variable for the property box
	global fontsize
	set font [.c itemcget $item -font]
	set fontsize [font config $font -size]
}

proc Deselect {} {
	RemoveHandles
	.c dtag selection
}

proc GetSelection {} {
	return [.c find withtag selection]
}

proc UpdateSelection {} {
	set item [GetSelection]
	if {$item == ""} {
		return
	}
	RemoveHandles
	CreateHandles $item
}

# ---- Menu bar

.c create rectangle 3 3 638 23 -fill lightgrey -tag menubar
.c create rectangle 300 4 380 22 -outline {} -fill lightgrey -tag menutitle

set HilitedItem ""

proc CreateMenu {} {
	global HilitedItem
	
	.c create rectangle 300 22 380 223 -fill lightgrey -tag menu
	for {set y 23} {$y < 223} {incr y 20} {
		.c create rectangle 301 $y 379 [expr $y + 18] -tag {menu menuitem} \
			-outline lightgrey -fill lightgrey
	}
	set HilitedItem ""
}

proc RemoveMenu {} {
	.c delete menu
	.c itemconfig menutitle -fill lightgrey
}

proc HiliteItem {item} {
	global HilitedItem ""
	if {$HilitedItem != ""} {
		.c itemconfig $HilitedItem -fill lightgrey
		set HilitedItem ""
	}
	if {$item != ""} {
		set HilitedItem $item
		.c itemconfig $HilitedItem -fill darkgrey
	}
}

# ---- Box within the canvas, with primitive window management

proc Box {path {title ""}} {
	set path .c$path
	if [winfo exist $path] {
		return ""
	}
	
	frame $path -bg black
	place $path -x 300 -y 220
	
	frame $path.box
	frame $path.close -bg lightgrey -width 15 -height 15
	label $path.title -text $title -fg white -bg black
	
	pack $path.box -side bottom -expand on -fill both -padx 2 -pady 2
	pack $path.close -side left -padx 6 -pady 2
	pack $path.title -side top -padx 2 -pady 2 -fill x
	
	StateMachine::enable closeBox $path.close
	StateMachine::enable dragBox $path
	StateMachine::enable dragBox $path.title
	
	return $path.box
}

proc CloseBox {path} {
	destroy [winfo parent $path]
}

StateMachine::define closeBox {
	state Start {
		when Down do {
			%W config -bg red
		} -> WaitUp
	}
	
	state WaitUp {
		when Leave do {
			%W config -bg lightgrey
		} -> Out
		when Up do {
			CloseBox %W
		} -> Start
	}

	state Out {
		when Enter do {
			%W config -bg red
		} -> WaitUp
		when Up do {
		} -> Start
	}
}

StateMachine::define dragBox {
	local P
	local box
	local boxpos
	
	state Start {
		when Down do {
			set P {%X %Y}
			set box %W
			while {[winfo parent $box]  != ".c"} {
				set box [winfo parent $box]
			}
			array set info [place info $box]
			set boxpos [list $info(-x) $info(-y)]
		} -> Drag
	}
	
	state Drag {
		when Move do {
			lset {x y} $P
			set dx [expr %X - $x]
			set dy [expr %Y - $y]
			lset {x y} $boxpos
			place config $box -x [expr $x + $dx] -y [expr $y + $dy]
		}
		when Up do {
		} -> Start
	}
}

# ---- Dialog Box

proc DboxChangeSize {} {
	set sel [GetSelection]
	if {$sel == ""} {
		return
	}
	set dbox [Box .dbox "Change Font Size"]
	if {$dbox == ""} {
		return
	}
	
	global done
	set done 0
	set font [.c itemcget $sel -font]
	label $dbox.label -text "New Font Size"
	entry $dbox.size -width 10
	$dbox.size insert 0 [font config $font -size]
	bind $dbox.size <Return> "$dbox.ok invoke"
	button $dbox.ok -text OK -command {set done 1} -default active
	
	pack $dbox.ok -side bottom -padx 5 -pady 5 -anchor se
	pack $dbox.size -side right -padx 5 -pady 5
	pack $dbox.label -side right -padx 5 -pady 5
	
	update idletasks
	focus $dbox.size
	$dbox.size select range 0 end
	
	vwait done
	# update fontsize for property box
	global fontsize
	set fontsize [$dbox.size get]
	font configure $font -size $fontsize

	# need to update so we get the handles properly positioned
	update idletasks
	UpdateSelection
	CloseBox $dbox
}

# ---- Property box

proc PboxChangeSize {} {
	set pbox [Box .pbox "Text properties"]
	if {$pbox == ""} {
		return
	}
	
	global fontsize
	label $pbox.label -text "Font Size"
	scale $pbox.size -from 6 -to 72 -orient horiz -resolution 1 -variable fontsize
	
	pack $pbox.label -side left -padx 5 -pady 5
	pack $pbox.size -side left -padx 5 -pady 5 -fill x
	
	trace variable fontsize w PboxSetFontSize
	
	bind $pbox.size <Delete> ClosePBox
}

proc PboxSetFontSize {args} {
	global fontsize
	set sel [GetSelection]
	if {$sel != ""} {
		set font [.c itemcget $sel -font]
		font config $font -size $fontsize
		update idletasks
		UpdateSelection
	}
}

proc ClosePbox {} {
	trace vdelete fontsize w PboxSetFontSize
}

# ---- Dragging handle

set startpos {0 0}
set startwidth 0
set startsize 0
set font ""

proc StartDrag {pos} {
	global startpos startwidth startsize fontsize font
	set startpos $pos
	set sel [GetSelection]
	set font [.c itemcget $sel -font]
	lset {x1 y1 x2 y2} [.c bbox $sel]
	set startwidth [expr $x2 - $x1]
	set startsize $fontsize
}

proc DoDrag {pos} {
	global startpos startwidth startsize fontsize font
	lset {sx sy} $startpos
	lset {x y} $pos
	set width [expr $startwidth + $x - $sx]
	set fontsize [expr int ($startsize * $width / $startwidth)] 
	font config $font -size $fontsize
	UpdateSelection
}

proc StopDrag {pos} {

}

# ---- State machine

StateMachine::define Canvas {
	state Start {
		# selecting a text item
		when Down onItem i withTag text do {
			Deselect
			Select $i
			arm 200
		} -> WaitDblClick
		
		# using the menubar
		when Enter onItem i withTag menutitle do {
			.c itemconfig menutitle -fill darkgrey
		}
		when Leave onItem i withTag menutitle do {
			.c itemconfig menutitle -fill lightgrey
		}
		when Down onItem i withTag menutitle do {
			# nothing - just catch event
		}
		when Up onItem i withTag menutitle do {
			CreateMenu
		} -> Menu
		
		# clicking a handle
		when Down at p onItem i withTag se do {
			StartDrag $p
		} -> DragHandle
		
		# clicking on the background
		when Down do {
			Deselect
			arm 200
		} -> WaitDblClick
	}
	
	# detect double clicks to open property box
	state WaitDblClick {
		when Down do {
			PboxChangeSize
		} -> Start
		when TimeOut -> Start
	}
	
	# drag handle 
	state DragHandle {
		when Move at p do {
			DoDrag $p
		}
		when Up at p do {
			StopDrag $p
		} -> Start
	}
	
	# select item in menu
	state Menu {
		when Enter onItem i withTag menuitem do {
			HiliteItem $i
		}
		when Leave onItem i withTag menuitem do {
			HiliteItem ""
		}
		when Up do {
			RemoveMenu
			global HilitedItem
			if {$HilitedItem != ""} {
				DboxChangeSize
			}
		} -> Start
	}
}

StateMachine::enable Canvas .c
